@@ -3,7 +3,7 @@ package YAHC;
 use strict;
 use warnings;
 
-our $VERSION = '0.017';
+our $VERSION = '0.018';
 
 use EV;
 use Time::HiRes;
@@ -12,6 +12,9 @@ use Scalar::Util qw/weaken/;
 use Fcntl qw/F_GETFL F_SETFL O_NONBLOCK/;
 use POSIX qw/EINPROGRESS EINTR EAGAIN EWOULDBLOCK strftime/;
 use Socket qw/PF_INET SOCK_STREAM $CRLF SOL_SOCKET SO_ERROR inet_aton inet_ntoa pack_sockaddr_in/;
+use constant SSL => $ENV{YAHC_NO_SSL} ? 0 : eval 'use IO::Socket::SSL 1.94 (); 1';
+use constant SSL_WANT_READ  => SSL ? IO::Socket::SSL::SSL_WANT_READ()  : 0;
+use constant SSL_WANT_WRITE => SSL ? IO::Socket::SSL::SSL_WANT_WRITE() : 0;
 
 sub YAHC::Error::NO_ERROR                () { 0 }
 sub YAHC::Error::REQUEST_TIMEOUT         () { 1 << 0 }
@@ -24,22 +27,29 @@ sub YAHC::Error::WRITE_ERROR             () { 1 << 12 }
 sub YAHC::Error::REQUEST_ERROR           () { 1 << 13 }
 sub YAHC::Error::RESPONSE_ERROR          () { 1 << 14 }
 sub YAHC::Error::CALLBACK_ERROR          () { 1 << 15 }
+sub YAHC::Error::SSL_ERROR               () { 1 << 16 }
 sub YAHC::Error::INTERNAL_ERROR          () { 1 << 31 }
 
-sub YAHC::State::INITIALIZED             () { 1 << 0 }
-sub YAHC::State::RESOLVE_DNS             () { 1 << 1 }
-sub YAHC::State::WAIT_SYNACK             () { 1 << 2 }
-sub YAHC::State::CONNECTED               () { 1 << 3 }
-sub YAHC::State::WRITING                 () { 1 << 4 }
-sub YAHC::State::READING                 () { 1 << 5 }
-sub YAHC::State::USER_ACTION             () { 1 << 15 }
-sub YAHC::State::COMPLETED               () { 1 << 30 } # terminal state
+sub YAHC::State::INITIALIZED             () { 0   }
+sub YAHC::State::RESOLVE_DNS             () { 5   }
+sub YAHC::State::WAIT_SYNACK             () { 10  }
+sub YAHC::State::CONNECTED               () { 15  }
+sub YAHC::State::SSL_HANDSHAKE           () { 20  }
+sub YAHC::State::WRITING                 () { 25  }
+sub YAHC::State::READING                 () { 30  }
+sub YAHC::State::USER_ACTION             () { 35  }
+sub YAHC::State::COMPLETED               () { 100 } # terminal state
 
 use constant {
-    HTTP_PORT                  => 80,
-    TCP_READ_CHUNK             => 65536,
-    CALLBACKS                  => [ qw/init_callback wait_synack_callback connected_callback
-                                       writing_callback reading_callback callback/ ],
+    # TCP_READ_CHUNK should *NOT* be lower than 16KB because of SSL things.
+    # https://metacpan.org/pod/distribution/IO-Socket-SSL/lib/IO/Socket/SSL.pod
+    # Another way might be if you try to sysread at least 16kByte all the time.
+    # 16kByte is the maximum size of an SSL frame and because sysread returns
+    # data from only a single SSL frame you can guarantee that there are no
+    # pending data.
+    TCP_READ_CHUNK              => 131072,
+    CALLBACKS                   => [ qw/init_callback wait_synack_callback connected_callback
+                                        writing_callback reading_callback callback/ ],
 };
 
 our @EXPORT_OK = qw/
@@ -72,52 +82,41 @@ sub new {
     # and more importantly to share index within the list
     $args->{_target} = _wrap_target_selection($args->{host}) if $args->{host};
 
-    # this's a radical way of avoiding circular references.
-    # let's see how it plays out in practise.
-    my %storage = (
-        watchers    => {},
-        callbacks   => {},
-        connections => {},
-    );
-
+    my %storage;
     my $self = bless {
         loop                => new EV::Loop,
         pid                 => $$, # store pid to detect forks
         storage             => \%storage,
-        watchers            => $storage{watchers},
-        callbacks           => $storage{callbacks},
-        connections         => $storage{connections},
         last_connection_id  => $$ * 1000,
-
         debug               => delete $args->{debug} || 0,
         keep_timeline       => delete $args->{keep_timeline} || 0,
         pool_args           => $args,
-        is_running          => 0,
     }, $class;
 
+    # this's a radical way of avoiding circular references.
+    # let's see how it plays out in practise.
     weaken($self->{storage});
-    weaken($self->{watchers});
-    weaken($self->{callbacks});
-    weaken($self->{connections});
+    weaken($self->{$_} = $storage{$_} = {}) for qw/watchers callbacks connections/;
     return $self, \%storage;
 }
 
 sub request {
     my ($self, @args) = @_;
-    my ($conn_id, $request);
     die 'YAHC: new_request() expects arguments' unless @args;
     die 'YAHC: storage object is destroyed' unless $self->{storage};
 
-    if (@args == 1) {
-        $conn_id = 'connection_' . $self->{last_connection_id}++;
-        $request = $args[0];
-    } else {
-        ($conn_id, $request) = @args;
-    }
-
+    my ($conn_id, $request) = (@args == 1 ? ('connection_' . $self->{last_connection_id}++, $args[0]) : @args);
     die "YAHC: Connection with name '$conn_id' already exists\n"
         if exists $self->{connections}{$conn_id};
 
+    my $pool_args = $self->{pool_args};
+    $request->{_target} = _wrap_target_selection($request->{host}) if $request->{host};
+    do { $request->{$_} ||= $pool_args->{$_} if $pool_args->{$_} } foreach (qw/host port scheme request_timeout
+                                                                               connect_timeout drain_timeout/);
+    my $scheme = $request->{scheme} ||= 'http';
+    die "YAHC: only support scheme http and https\n" unless $scheme eq 'http' || $scheme eq 'https';
+    die "YAHC: host must be defined\n" unless $request->{host};
+
     my $conn = {
         id          => $conn_id,
         request     => $request,
@@ -130,17 +129,9 @@ sub request {
         selected_target => [],
     };
 
-    my $pool_args = $self->{pool_args};
-    $request->{_target} = _wrap_target_selection($request->{host}) if $request->{host};
-    do { $request->{$_} ||= $pool_args->{$_} if $pool_args->{$_} } foreach (qw/host port scheme request_timeout
-                                                                               connect_timeout drain_timeout/);
-    $request->{scheme} ||= 'http';
-    die "YAHC: only support scheme http\n" unless $request->{scheme} eq 'http';
-    die "YAHC: host must be defined\n" unless $request->{host};
-
     my %callbacks;
     foreach (@{ CALLBACKS() }) {
-        next unless $request->{$_};
+        next unless exists $request->{$_};
         $callbacks{$_} = delete $request->{$_};
         $conn->{"has_$_"} = 1;
     }
@@ -164,7 +155,6 @@ sub request {
 sub drop {
     my ($self, $c) = @_;
     my $conn_id = ref($c) eq 'HASH' ? $c->{id} : $c;
-
     my $conn = $self->{connections}{$conn_id}
       or return;
 
@@ -178,7 +168,7 @@ sub drop {
 sub run         { shift->_run(0, @_)            }
 sub run_once    { shift->_run(EV::RUN_ONCE)     }
 sub run_tick    { shift->_run(EV::RUN_NOWAIT)   }
-sub is_running  { shift->{is_running}           }
+sub is_running  { !!shift->{loop}->depth        }
 sub loop        { shift->{loop}                 }
 
 ################################################################################
@@ -217,7 +207,7 @@ sub yahc_conn_target {
     my $target = $_[0]->{selected_target};
     return unless $target && scalar @{ $target };
     my ($host, $ip, $port) = @{ $target };
-    return ($host || $ip) . ($port ne "80" ? ":$port" : '');
+    return ($host || $ip) . ($port ne '80' && $port ne '443' ? ":$port" : '');
 }
 
 sub yahc_conn_url {
@@ -225,10 +215,10 @@ sub yahc_conn_url {
     my $request = $_[0]->{request};
     return unless $target && @{ $target };
 
-    my ($host, $ip, $port) = @{ $target };
-    return $request->{scheme} . "://"
+    my ($host, $ip, $port, $scheme) = @{ $target };
+    return "$scheme://"
            . ($host || $ip)
-           . ($port ne "80" ? ":$port" : '')
+           . ($port ne '80' && $port ne '443' ? ":$port" : '')
            . ($request->{path} || "/")
            . (defined $request->{query_string} ? ("?" . $request->{query_string}) : "");
 }
@@ -240,8 +230,7 @@ sub yahc_conn_url {
 sub _run {
     my ($self, $how, $until_state, @cs) = @_;
     die "YAHC: storage object is destroyed\n" unless $self->{storage};
-    die "YAHC: reentering run\n" if $self->{is_running};
-    $self->{is_running} = 1;
+    die "YAHC: reentering run\n" if $self->{loop}->depth;
 
     if ($self->{pid} != $$) {
         _log_message('YAHC: reinitializing event loop after forking') if $self->{debug};
@@ -278,8 +267,6 @@ sub _run {
     } else {
         $loop->run($how || 0);
     }
-
-    $self->{is_running} = 0;
 }
 
 sub _break {
@@ -313,20 +300,19 @@ sub _check_stop_condition {
 sub _set_init_state {
     my ($self, $conn_id) = @_;
 
-    my $conn = $self->{connections}{$conn_id};
-    my $watchers = $self->{watchers}{$conn_id};
-    return $self->_set_completed_state($conn_id) unless $conn && $watchers;
+    my $conn = $self->{connections}{$conn_id}  or die "YAHC: unknown connection id $conn_id\n";
+    my $watchers = $self->{watchers}{$conn_id} or die "YAHC: no watchers for connection id $conn_id\n";
 
     $conn->{response} = { status => 0 };
     $conn->{state} = YAHC::State::INITIALIZED();
     _register_in_timeline($conn, "new state %s", _strstate($conn->{state})) if $conn->{keep_timeline};
-    $self->_call_state_callback($conn, 'init_callback') if $conn->{has_init_callback};
+    $self->_call_state_callback($conn, 'init_callback') if exists $conn->{has_init_callback};
 
     my $continue = 1;
     while ($continue) {
         delete $watchers->{io}; # implicit stop
         my $fh = delete $watchers->{_fh};
-        $fh && shutdown($fh, 2), undef $fh;
+        $fh && close($fh), undef $fh;
 
         my $attempt = $conn->{attempt}++;
         if ($attempt > $conn->{retries}) {
@@ -338,14 +324,13 @@ sub _set_init_state {
         # I can set them here under $attempt == 0 condition.
         # Also, in _set_until_state_timer I would go to USER_ACTION_STATE
         # instead of reinitializing connection.
-
         $self->_set_request_timer($conn_id)    if $conn->{request}{request_timeout};
         $self->_set_connection_timer($conn_id) if $conn->{request}{connect_timeout};
         $self->_set_drain_timer($conn_id)      if $conn->{request}{drain_timeout};
 
         eval {
-            my ($host, $ip, $port) = _get_next_target($conn);
-            _register_in_timeline($conn, "Target $host:$port ($ip:$port) chosen for attempt #$attempt") if $conn->{keep_timeline};
+            my ($host, $ip, $port, $scheme) = _get_next_target($conn);
+            _register_in_timeline($conn, "Target $scheme://$host:$port ($ip:$port) chosen for attempt #$attempt") if $conn->{keep_timeline};
 
             my $sock = _build_socket_and_connect($ip, $port, $conn->{request});
             $self->_set_wait_synack_state($conn_id, $sock);
@@ -354,8 +339,6 @@ sub _set_init_state {
         } or do {
             my $error = $@ || 'zombie error';
             _register_error($conn, YAHC::Error::CONNECT_ERROR(), "Connection attempt failed: $error");
-            #$self->_set_set_init_state_timer($conn_id); # trigger background timer which will do reconnection attempt # TODO
-            #$continue = 0;
         };
     }
 }
@@ -363,16 +346,14 @@ sub _set_init_state {
 sub _set_wait_synack_state {
     my ($self, $conn_id, $sock) = @_;
 
-    my $conn = $self->{connections}{$conn_id};
-    my $watchers = $self->{watchers}{$conn_id};
-    return $self->_set_completed_state($conn_id) unless $conn && $watchers;
-    _assert_state($conn, YAHC::State::INITIALIZED()) if $conn->{debug};
+    my $conn = $self->{connections}{$conn_id}  or die "YAHC: unknown connection id $conn_id\n";
+    my $watchers = $self->{watchers}{$conn_id} or die "YAHC: no watchers for connection id $conn_id\n";
 
     $conn->{state} = YAHC::State::WAIT_SYNACK();
     _register_in_timeline($conn, "new state %s", _strstate($conn->{state})) if $conn->{keep_timeline};
-    $self->_call_state_callback($conn, 'wait_synack_callback') if $conn->{has_wait_synack_callback};
+    $self->_call_state_callback($conn, 'wait_synack_callback') if exists $conn->{has_wait_synack_callback};
 
-    my $wait_synack_cb = sub {
+    my $wait_synack_cb = $self->_get_safe_wrapper($conn, sub {
         my $sockopt = getsockopt($sock, SOL_SOCKET, SO_ERROR);
         if (!$sockopt) {
             _register_error($conn, YAHC::Error::CONNECT_ERROR(), "Failed to do getsockopt(): $!");
@@ -389,84 +370,159 @@ sub _set_wait_synack_state {
 
         $conn->{state} = YAHC::State::CONNECTED();
         _register_in_timeline($conn, "new state %s", _strstate($conn->{state})) if $conn->{keep_timeline};
-        $self->_call_state_callback($conn, 'connected_callback') if $conn->{has_connected_callback};
-        return if $self->{stop_condition} && $self->_check_stop_condition($conn);
+        $self->_call_state_callback($conn, 'connected_callback') if exists $conn->{has_connected_callback};
+        return if exists $self->{stop_condition} && $self->_check_stop_condition($conn);
 
+        return $self->_set_ssl_handshake_state($conn_id) if $conn->{is_ssl};
         $self->_set_write_state($conn_id);
-    };
+    });
 
     $watchers->{_fh} = $sock;
     $watchers->{io} = $self->{loop}->io($sock, EV::WRITE, $wait_synack_cb);
+    $self->_check_stop_condition($conn) if exists $self->{stop_condition};
+}
+
+sub _set_ssl_handshake_state {
+    my ($self, $conn_id) = @_;
+
+    my $conn = $self->{connections}{$conn_id}  or die "YAHC: unknown connection id $conn_id\n";
+    my $watchers = $self->{watchers}{$conn_id} or die "YAHC: no watchers for connection id $conn_id\n";
+
+    $conn->{state} = YAHC::State::SSL_HANDSHAKE();
+    _register_in_timeline($conn, "new state %s", _strstate($conn->{state})) if $conn->{keep_timeline};
+    #$self->_call_state_callback($conn, 'writing_callback') if $conn->{has_writing_callback}; TODO
+
+    my $fh = $watchers->{_fh};
+    my $hostname = $conn->{selected_target}[0];
+
+    my %options = (
+        SSL_verifycn_name => $hostname,
+        IO::Socket::SSL->can_client_sni ? ( SSL_hostname => $hostname ) : (),
+        %{ $conn->{request}{ssl_options} || {} },
+    );
+
+    if ($conn->{keep_timeline}) {
+        my $options_msg = join(', ', map { "$_=" . ($options{$_} || '') } keys %options);
+        _register_in_timeline($conn, "start SSL handshake with options: $options_msg");
+    }
+
+    if (!IO::Socket::SSL->start_SSL($fh, %options, SSL_startHandshake => 0)) {
+        _register_error($conn, YAHC::Error::SSL_ERROR(), "Failed to start SSL session: $IO::Socket::SSL::SSL_ERROR");
+        return $self->_set_completed_state($conn_id);
+    }
+
+    my $handshake_cb = $self->_get_safe_wrapper($conn, sub {
+        my $w = shift;
+        if ($fh->connect_SSL) {
+            _register_in_timeline($conn, "SSL handshake successfully completed") if $conn->{keep_timeline};
+            return $self->_set_write_state($conn_id);
+        }
+
+        if ($! == EWOULDBLOCK) {
+            return $w->events(EV::READ)  if $IO::Socket::SSL::SSL_ERROR == SSL_WANT_READ;
+            return $w->events(EV::WRITE) if $IO::Socket::SSL::SSL_ERROR == SSL_WANT_WRITE;
+        }
+
+        _register_error($conn, YAHC::Error::SSL_ERROR(), "Failed to complete SSL handshake: <$!> SSL_ERROR: <$IO::Socket::SSL::SSL_ERROR>");
+        $self->_set_init_state($conn_id);
+    });
+
+    my $watcher = $watchers->{io};
+    $watcher->cb($handshake_cb);
+    $watcher->events(EV::WRITE | EV::READ);
     $self->_check_stop_condition($conn) if $self->{stop_condition};
 }
 
 sub _set_write_state {
     my ($self, $conn_id) = @_;
 
-    my $conn = $self->{connections}{$conn_id};
-    my $watchers = $self->{watchers}{$conn_id};
-    my $watcher = $watchers->{io};
-    return $self->_set_completed_state($conn_id) unless $conn && $watchers && $watcher;
-    _assert_connected($conn) if $conn->{debug};
+    my $conn = $self->{connections}{$conn_id}  or die "YAHC: unknown connection id $conn_id\n";
+    my $watchers = $self->{watchers}{$conn_id} or die "YAHC: no watchers for connection id $conn_id\n";
 
     $conn->{state} = YAHC::State::WRITING();
     _register_in_timeline($conn, "new state %s", _strstate($conn->{state})) if $conn->{keep_timeline};
-    $self->_call_state_callback($conn, 'writing_callback') if $conn->{has_writing_callback};
+    $self->_call_state_callback($conn, 'writing_callback') if exists $conn->{has_writing_callback};
 
-    my $fd = fileno($watchers->{_fh});
+    my $fh = $watchers->{_fh};
     my $buf = _build_http_message($conn);
     my $length = length($buf);
 
     _register_in_timeline($conn, "sending request of $length bytes") if $conn->{keep_timeline};
 
-    my $write_cb = sub {
-        my $wlen = POSIX::write($fd, $buf, $length);
+    my $write_cb = $self->_get_safe_wrapper($conn, sub {
+        my $w = shift;
+        my $wlen = syswrite($fh, $buf, $length);
 
-        if (!defined $wlen || $wlen == 0) {
-            return if $! == EWOULDBLOCK || $! == EAGAIN || $! == EINTR;
-            _register_error($conn, YAHC::Error::WRITE_ERROR(), "Failed to send TCP data: $!");
+        if (!defined $wlen) {
+            if ($conn->{is_ssl}) {
+                if ($! == EWOULDBLOCK) {
+                    return $w->events(EV::READ)  if $IO::Socket::SSL::SSL_ERROR == SSL_WANT_READ;
+                    return $w->events(EV::WRITE) if $IO::Socket::SSL::SSL_ERROR == SSL_WANT_WRITE;
+                }
+
+                _register_error($conn, YAHC::Error::WRITE_ERROR() | YAHC::Error::SSL_ERROR(),
+                                "Failed to send HTTPS data: <$!> SSL_ERROR: <$IO::Socket::SSL::SSL_ERROR>");
+                return $self->_set_init_state($conn_id);
+            }
+
+            return if $! == EWOULDBLOCK || $! == EINTR || $! == EAGAIN;
+            _register_error($conn, YAHC::Error::WRITE_ERROR(), "Failed to send HTTP data: $!");
+            $self->_set_init_state($conn_id);
+        } elsif ($wlen == 0) {
+            _register_error($conn, YAHC::Error::WRITE_ERROR(), "syswrite returned 0");
             $self->_set_init_state($conn_id);
         } else {
             substr($buf, 0, $wlen, '');
             $length -= $wlen;
             $self->_set_read_state($conn_id) if $length == 0;
         }
-    };
+    });
 
+    my $watcher = $watchers->{io};
     $watcher->cb($write_cb);
     $watcher->events(EV::WRITE);
-    $self->_check_stop_condition($conn) if $self->{stop_condition};
+    $self->_check_stop_condition($conn) if exists $self->{stop_condition};
 }
 
 sub _set_read_state {
     my ($self, $conn_id) = @_;
 
-    my $conn = $self->{connections}{$conn_id};
-    my $watchers = $self->{watchers}{$conn_id};
-    my $watcher = $watchers->{io};
-    return $self->_set_completed_state($conn_id) unless $conn && $watchers && $watcher;
-    _assert_connected($conn) if $conn->{debug};
+    my $conn = $self->{connections}{$conn_id}  or die "YAHC: unknown connection id $conn_id\n";
+    my $watchers = $self->{watchers}{$conn_id} or die "YAHC: no watchers for connection id $conn_id\n";
 
     $conn->{state} = YAHC::State::READING();
     _register_in_timeline($conn, "new state %s", _strstate($conn->{state})) if $conn->{keep_timeline};
-    $self->_call_state_callback($conn, 'reading_callback') if $conn->{has_reading_callback};
+    $self->_call_state_callback($conn, 'reading_callback') if exists $conn->{has_reading_callback};
 
     my $buf = '';
     my $neck_pos = 0;
     my $decapitated = 0;
     my $content_length = 0;
-    my $fd = fileno($watchers->{_fh});
+    my $fh = $watchers->{_fh};
 
-    my $read_cb = sub {
-        my $rlen = POSIX::read($fd, my $b = '', TCP_READ_CHUNK);
+    my $read_cb = $self->_get_safe_wrapper($conn, sub {
+        my $w = shift;
+        my $rlen = sysread($fh, my $b = '', TCP_READ_CHUNK);
 
-        if (!defined $rlen || $rlen == 0) {
-            return if $! == EWOULDBLOCK || $! == EAGAIN || $! == EINTR;
-            if (not defined $rlen) {
-                _register_error($conn, YAHC::Error::READ_ERROR(), "Failed to receive TCP data: $!");
-            } elsif ($content_length > 0) { # i.e. rlen == 0 and $content_length > 0
+        if (!defined $rlen) {
+            if ($conn->{is_ssl}) {
+                if ($! == EWOULDBLOCK) {
+                    return $w->events(EV::READ)  if $IO::Socket::SSL::SSL_ERROR == SSL_WANT_READ;
+                    return $w->events(EV::WRITE) if $IO::Socket::SSL::SSL_ERROR == SSL_WANT_WRITE;
+                }
+
+                _register_error($conn, YAHC::Error::READ_ERROR() | YAHC::Error::SSL_ERROR(),
+                                "Failed to receive HTTPS data: <$!> SSL_ERROR: <$IO::Socket::SSL::SSL_ERROR>");
+                return $self->_set_init_state($conn_id);
+            }
+
+            return if $! == EWOULDBLOCK || $! == EINTR || $! == EAGAIN;
+            _register_error($conn, YAHC::Error::READ_ERROR(), "Failed to receive HTTP data: $!");
+            $self->_set_init_state($conn_id);
+        } elsif ($rlen == 0) {
+            if ($content_length > 0) {
                 _register_error($conn, YAHC::Error::READ_ERROR(), "Premature EOF, expect %d bytes more", $content_length - length($buf));
-            } else { # i.e. rlen == 0
+            } else {
                 _register_error($conn, YAHC::Error::READ_ERROR(), "Premature EOF");
             }
 
@@ -493,11 +549,12 @@ sub _set_read_state {
                 $self->_set_user_action_state($conn_id);
             }
         }
-    };
+    });
 
+    my $watcher = $watchers->{io};
     $watcher->cb($read_cb);
     $watcher->events(EV::READ);
-    $self->_check_stop_condition($conn) if $self->{stop_condition};
+    $self->_check_stop_condition($conn) if exists $self->{stop_condition};
 }
 
 sub _set_user_action_state {
@@ -515,7 +572,7 @@ sub _set_user_action_state {
     _register_in_timeline($conn, "new state %s", _strstate($conn->{state})) if $conn->{keep_timeline};
     _register_error($conn, $error, $strerror) if $error;
 
-    return $self->_set_completed_state($conn_id) unless $conn->{has_callback};
+    return $self->_set_completed_state($conn_id) unless exists $conn->{has_callback};
     my $cb = $self->{callbacks}{$conn_id}{callback};
 
     eval {
@@ -559,10 +616,10 @@ sub _set_completed_state {
     _register_in_timeline($conn, "new state %s", _strstate($conn->{state})) if $conn->{keep_timeline};
 
     my $fh = $watchers->{_fh};
-    $fh && shutdown($fh, 2);
+    $fh && close($fh);
     undef $watchers; # implicit stop
 
-    $self->_check_stop_condition($conn) if $self->{stop_condition};
+    $self->_check_stop_condition($conn) if exists $self->{stop_condition};
 }
 
 sub _build_socket_and_connect {
@@ -585,15 +642,17 @@ sub _build_socket_and_connect {
 
 sub _get_next_target {
     my $conn = shift;
-    my ($host, $ip, $port) = $conn->{request}{_target}->($conn);
+    my ($host, $ip, $port, $scheme) = $conn->{request}{_target}->($conn);
 
     # TODO STATE_RESOLVE_DNS
     ($host, $port) = ($1, $2) if !$port && $host =~ m/^(.+):([0-9]+)$/o;
     $ip = $host if !$ip && $host =~ m/^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$/o;
     $ip ||= inet_ntoa(gethostbyname($host) or die "Failed to resolve '$host': $!\n");
-    $port ||= $conn->{request}{port} || HTTP_PORT;
+    $scheme ||= $conn->{request}{scheme} || 'http';
+    $port   ||= $conn->{request}{port} || ($scheme eq 'https' ? 443 : 80);
 
-    return @{ $conn->{selected_target} = [ $host, $ip, $port ] };
+    $conn->{is_ssl} = $scheme eq 'https';
+    return @{ $conn->{selected_target} = [ $host, $ip, $port, $scheme ] };
 }
 
 ################################################################################
@@ -608,15 +667,14 @@ sub _set_until_state_timer {
     my ($self, $conn_id, $timeout_name, $state, $error_to_report) = @_;
 
     my $timer_name = $timeout_name . '_timer';
-    my $conn = $self->{connections}{$conn_id};
-    my $watchers = $self->{watchers}{$conn_id};
-    return $self->_set_completed_state($conn_id) unless $conn && $watchers;
+    my $conn = $self->{connections}{$conn_id}  or die "YAHC: unknown connection id $conn_id\n";
+    my $watchers = $self->{watchers}{$conn_id} or die "YAHC: no watchers for connection id $conn_id\n";
 
     delete $watchers->{$timer_name}; # implicit stop
     my $timeout = $conn->{request}{$timeout_name};
     return unless $timeout;
 
-    my $timer_cb = sub {
+    my $timer_cb = sub { # there is nothing what can throw exception
         if ($conn->{state} < $state) {
             _register_error($conn, $error_to_report, "$timeout_name of %.3fs expired", $timeout);
             $self->_set_init_state($conn_id);
@@ -626,6 +684,8 @@ sub _set_until_state_timer {
     };
 
     _register_in_timeline($conn, "setting $timeout_name to %.3fs", $timeout) if $conn->{keep_timeline};
+
+    $self->{loop}->now_update;
     my $w = $watchers->{$timer_name} = $self->{loop}->timer_ns($timeout, 0, $timer_cb);
     $w->priority(-2); # set lowest priority
     $w->start;
@@ -710,10 +770,22 @@ sub _call_state_callback {
     # $self->{loop}->now_update; # XXX expect state callbacks to be small
 }
 
+sub _get_safe_wrapper {
+    my ($self, $conn, $sub) = @_;
+    return sub { eval {
+        $sub->(@_);
+        1;
+    } or do {
+        my $error = $@ || 'zombie error';
+        _register_error($conn, YAHC::Error::INTERNAL_ERROR(), "Exception callback: $error");
+        warn "YAHC: exception in callback: $error";
+        $self->_set_completed_state($conn->{id});
+    }};
+}
+
 sub _register_in_timeline {
     my ($conn, $format, @arguments) = @_;
     my $event = sprintf("$format", @arguments);
-    $event =~ s/\s+$//g;
     _log_message("YAHC connection '%s': %s", $conn->{id}, $event) if $conn->{debug};
     push @{ $conn->{timeline} ||= [] }, [ $event, $conn->{state}, Time::HiRes::time ] if $conn->{keep_timeline};
 }
@@ -721,25 +793,10 @@ sub _register_in_timeline {
 sub _register_error {
     my ($conn, $error, $format, @arguments) = @_;
     my $strerror = sprintf("$format", @arguments);
-    $strerror =~ s/\s+$//g;
-    _register_in_timeline($conn, "error=$strerror (errno=$error)") if $conn->{keep_timeline};
+    _register_in_timeline($conn, "error=$strerror ($error)") if $conn->{keep_timeline};
     push @{ $conn->{errors} ||= [] }, [ $error, $strerror, [ @{ $conn->{selected_target} } ], Time::HiRes::time ];
 }
 
-sub _assert_state {
-    my ($conn, $expected_state) = @_;
-    return $conn->{state} == $expected_state;
-    die sprintf("YAHC connection '%s' is in unexpected state %s, expected %s\n",
-                $conn->{id}, _strstate($conn->{state}), _strstate($expected_state));
-}
-
-sub _assert_connected {
-    my $conn = shift;
-    return if $conn->{state} >= YAHC::State::CONNECTED() && $conn->{state} < YAHC::State::COMPLETED();
-    die sprintf("YAHC connection '%s' expected to be in connected state, but it's in %s\n",
-                $conn->{id}, _strstate($conn->{state}));
-}
-
 sub _strstate {
     my $state = shift;
     return 'STATE_INIT'         if $state eq YAHC::State::INITIALIZED();
@@ -748,6 +805,7 @@ sub _strstate {
     return 'STATE_CONNECTED'    if $state eq YAHC::State::CONNECTED();
     return 'STATE_WRITING'      if $state eq YAHC::State::WRITING();
     return 'STATE_READING'      if $state eq YAHC::State::READING();
+    return 'STATE_SSL_HANDSHAKE'if $state eq YAHC::State::SSL_HANDSHAKE();
     return 'STATE_USER_ACTION'  if $state eq YAHC::State::USER_ACTION();
     return 'STATE_COMPLETED'    if $state eq YAHC::State::COMPLETED();
     return "<unknown state $state>";
@@ -757,7 +815,7 @@ sub _log_message {
     my $format = shift;
     my $now = Time::HiRes::time;
     my ($sec, $ms) = split(/[.]/, $now);
-    printf STDERR "[%s.%05d] [$$] $format\n", POSIX::strftime('%F %T', localtime($now)), $ms || 0, @_;
+    printf STDERR "[%s.%-5d] [$$] $format\n", POSIX::strftime('%F %T', localtime($now)), $ms || 0, @_;
 }
 
 1;
